unit DropMetaSource;

// -----------------------------------------------------------------------------
// Project:         Drag and Drop Component Suite
// Component Names: TDropMetaSource
// Module:          DropMetaSource
// Description:     Implements Dragging & Dropping of Metafiles
//                  FROM your application to another.
// Version:         3.7
// Date:            01-AUG-2010
// Target:          Win32, Delphi 3 - Delphi 5, C++ Builder 3, C++ Builder 4
// Authors:         Angus Johnson,   ajohnson@rpi.net.au
//                  Anders Melander, anders@melander.dk
//                                   http://www.melander.dk
//                  seizo,           seizo@mars.dti.ne.jp
// Copyright        (C) 1997-99 Angus Johnson & Anders Melander
// Copyright        (C) 2011 seizo,  seizo@mars.dti.ne.jp
// -----------------------------------------------------------------------------

interface

uses
  Windows, Classes, SysUtils, Graphics, ClipBrd, ActiveX, DropSource;

type
  TDropMetaSource = class(TDropSource)
  private
    FMetafile: TMetafile;
    FMfCanvas: TMetafileCanvas;
    FFormatTymed: String;
    procedure SetMetafile(const Value: TMetafile);
  protected
    function DoGetData(const FormatEtcIn: TFormatEtc; out Medium: TStgMedium):HResult; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function CutOrCopyToClipboard: Boolean; override;
    property FormatTymed: String read FFormatTymed;
  published
    property Metafile: TMetafile read FMetafile write SetMetafile;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('DragDrop', [TDropMetaSource]);
end;

// -----------------------------------------------------------------------------
//			TDropMetaSource
// -----------------------------------------------------------------------------

constructor TDropMetaSource.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FMetafile := TMetafile.Create;
  FMetafile.Width := 100;
  FMetafile.Height := 100;
  { Prepare Metafile Canvas }
  FMfCanvas := TMetafileCanvas.Create(FMetafile, GetDC(0));
  try
    { Set Pen and Brush of FMfCanvas }
    FMfCanvas.Pen.Color := clBtnFace;
    FMfCanvas.Pen.Style := psSolid;
    FMfCanvas.Brush.Color := clBtnFace;
    FMfCanvas.Brush.Style := bsSolid;
    { Draw something }
    FMfCanvas.FillRect(Rect(0, 0, FMetafile.Width, FMetafile.Height));
  finally
    FMfCanvas.Free;        // First drawn open
  end;

  DragTypes := [dtCopy]; // Default to Copy
  FFormatTymed := '::';

  AddFormatEtc(CF_METAFILEPICT, Nil, DVASPECT_CONTENT, -1, TYMED_MFPICT);
  AddFormatEtc(CF_ENHMETAFILE, Nil, DVASPECT_CONTENT, -1, TYMED_ENHMF);
  AddFormatEtc(CF_METAFILEPICT, Nil, DVASPECT_CONTENT, -1, TYMED_HGLOBAL);
  AddFormatEtc(CF_ENHMETAFILE, Nil, DVASPECT_CONTENT, -1, TYMED_HGLOBAL);
end;
// ----------------------------------------------------------------------------- 

destructor TDropMetaSource.destroy;
begin
  FMetafile.Free;
  inherited Destroy;
end;
// ----------------------------------------------------------------------------- 

procedure TDropMetaSource.SetMetafile(const Value: TMetafile);
begin
  FMetafile.Assign(Value);
end;
// ----------------------------------------------------------------------------- 

function TDropMetaSource.CutOrCopyToClipboard: Boolean;
var
  MFormat: Word;
  AData: THandle;
  APalette: HPALETTE;
begin
  Result := False;
  if FMetafile.Empty then Exit;
  FMetafile.SaveToClipboardFormat(MFormat, AData, APalette);
  try
    if AData = 0 then Exit;
    Clipboard.SetAsHandle(MFormat, AData);
    Result := True;
  except
    raise Exception.Create('Unable to copy Metafile to Clipboard.');
  end;
end;
// ----------------------------------------------------------------------------- 

function TDropMetaSource.DoGetData(const FormatEtcIn: TFormatEtc; out Medium: TStgMedium):HResult;
var
  fmt: WORD;
  pal: HPALETTE;
begin

  Medium.tymed := 0;
  Medium.UnkForRelease := Nil;
  Medium.hGlobal := 0;
  FFormatTymed := '::';
  //--------------------------------------------------------------------------
  if not FMetafile.Empty and
     (FormatEtcIn.cfFormat = CF_ENHMETAFILE) and
     (FormatEtcIn.dwAspect = DVASPECT_CONTENT) and
     (FormatEtcIn.tymed and TYMED_ENHMF <> 0) then
  begin
    FFormatTymed := 'CF_ENHMETAFILE::TYMED_ENHMF';
    try
      //This next line just gets a copy of the metafile handle...
      FMetafile.SaveToClipboardFormat(fmt, THandle(Medium.hEnhMetaFile), pal);
      if pal <> 0 then
        DeleteObject(pal);
      if Medium.hEnhMetaFile <> 0 then
      begin
        Medium.tymed := TYMED_ENHMF;
        Result := S_OK;
      end
      else
        Result := E_OUTOFMEMORY;
    except
      Result := E_OUTOFMEMORY;
    end;
  end
  //--------------------------------------------------------------------------
  else if not FMetafile.Empty and
     (FormatEtcIn.cfFormat = CF_METAFILEPICT) and
     (FormatEtcIn.dwAspect = DVASPECT_CONTENT) and
     (FormatEtcIn.tymed and TYMED_MFPICT <> 0) then
  begin
    FFormatTymed := 'CF_METAFILEPICT::TYMED_MFPICT';
    try
      //This next line just gets a copy of the metafile handle...
      FMetafile.SaveToClipboardFormat(fmt, THandle(Medium.hMetafilePict), pal);
      if pal <> 0 then
        DeleteObject(pal);
      if Medium.hMetafilePict <> 0 then
      begin
        Medium.tymed := TYMED_MFPICT;
        Result := S_OK;
      end
      else
        Result := E_OUTOFMEMORY;
    except
      Result := E_OUTOFMEMORY;
    end;
  end
  //--------------------------------------------------------------------------
  else if not FMetafile.Empty and
     (FormatEtcIn.cfFormat = CF_METAFILEPICT) and
     (FormatEtcIn.dwAspect = DVASPECT_CONTENT) and
     (FormatEtcIn.tymed and TYMED_HGLOBAL <> 0) then
  begin
    FFormatTymed := 'CF_METAFILEPICT::TYMED_HGLOBAL';
    try
      //This next line just gets a copy of the metafile handle...
      FMetafile.SaveToClipboardFormat(fmt, THandle(Medium.hGlobal), pal);
      if pal <> 0 then
        DeleteObject(pal);
      if Medium.hGlobal <> 0 then
      begin
        Medium.tymed := TYMED_HGLOBAL;
        Result := S_OK;
      end
      else
        Result := E_OUTOFMEMORY;
    except
      Result := E_OUTOFMEMORY;
    end;
  end
  //--------------------------------------------------------------------------
  else if not FMetafile.Empty and
     (FormatEtcIn.cfFormat = CF_ENHMETAFILE) and
     (FormatEtcIn.dwAspect = DVASPECT_CONTENT) and
     (FormatEtcIn.tymed and TYMED_HGLOBAL <> 0) then
  begin
    FFormatTymed := 'CF_ENHMETAFILE::TYMED_HGLOBAL';
    try
      //This next line just gets a copy of the metafile handle...
      FMetafile.SaveToClipboardFormat(fmt, THandle(Medium.hGlobal), pal);
      if pal <> 0 then
        DeleteObject(pal);
      if Medium.hGlobal <> 0 then
      begin
        Medium.tymed := TYMED_HGLOBAL;
        Result := S_OK;
      end
      else
        Result := E_OUTOFMEMORY;
    except
      Result := E_OUTOFMEMORY;
    end;
  end
  //--------------------------------------------------------------------------
  else
    Result := DV_E_FORMATETC;
end;
// -----------------------------------------------------------------------------

end.

