unit DropMetaTarget;

// -----------------------------------------------------------------------------
// Project:         Drag and Drop Component Suite
// Component Names: TDropMetaTarget
// Module:          DropMetaTarget
// Description:     Implements Dragging & Dropping of Metafiles
//                  FROM your application to another.
// Version:         3.7
// Date:            22-JUL-1999
// Target:          Win32, Delphi 3 - Delphi 5, C++ Builder 3, C++ Builder 4
// Authors:         Angus Johnson,   ajohnson@rpi.net.au
//                  Anders Melander, anders@melander.dk
//                                   http://www.melander.dk
//                  seizo,           seizo@mars.dti.ne.jp
// Copyright        (C) 1997-99 Angus Johnson & Anders Melander
// Copyright        (C) 2011 seizo,  seizo@mars.dti.ne.jp
// -----------------------------------------------------------------------------

interface

uses
  Windows, Classes, SysUtils, Graphics, ActiveX, DropSource, DropTarget;

type
  TDropMetaTarget = class(TDropTarget)
  private
    FMetafile: TMetafile;
    FFormatTymed: String;
  protected
    procedure ClearData; override;
    function DoGetData: Boolean; override;
    function HasValidFormats: Boolean; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Metafile: TMetafile read FMetafile;
    property FormatTymed: String read FFormatTymed;
  end;

procedure Register;
function WMF2EMF(const Mfp: TMetaFilePict): hEnhMetaFile;

implementation

const
  MetaFormatEtc: TFormatEtc = (cfFormat: CF_METAFILEPICT;
    ptd: Nil; dwAspect: DVASPECT_CONTENT; lindex: -1; tymed: TYMED_MFPICT);
  EnhFormatEtc: TFormatEtc = (cfFormat: CF_ENHMETAFILE;
    ptd: Nil; dwAspect: DVASPECT_CONTENT; lindex: -1; tymed: TYMED_ENHMF);
  MetaFormatEtc2: TFormatEtc = (cfFormat: CF_METAFILEPICT;
    ptd: Nil; dwAspect: DVASPECT_CONTENT; lindex: -1; tymed: TYMED_HGLOBAL);
  EnhFormatEtc2: TFormatEtc = (cfFormat: CF_ENHMETAFILE;
    ptd: Nil; dwAspect: DVASPECT_CONTENT; lindex: -1; tymed: TYMED_HGLOBAL);

procedure Register;
begin
  RegisterComponents('DragDrop', [TDropMetaTarget]);
end;

// -----------------------------------------------------------------------------
//	Miscellaneous Metafile Function
// -----------------------------------------------------------------------------

function WMF2EMF(const Mfp: TMetaFilePict): hEnhMetaFile;
var
  Bits: Pointer;
  Length: UINT;
  RefDC: HDC;
begin
  Length := GetMetaFileBitsEx(Mfp.hMF, 0, Nil);
  if Length = 0 then
    RaiseLastWin32Error;
  GetMem(Bits, Length);
  try
    if GetMetaFileBitsEx(Mfp.hMF, Length, Bits) < Length then
      RaiseLastWin32Error;
    RefDC := GetDC(0);
    try
        Result := SetWinMetaFileBits(Length, Bits, RefDC, Mfp);
    finally
      ReleaseDC(0, RefDC);
    end;
    if Result = 0 then
      RaiseLastWin32Error;

  finally
    FreeMem(Bits);
  end;
end;
// ----------------------------------------------------------------------------- 

// -----------------------------------------------------------------------------
//			TDropMetaTarget
// -----------------------------------------------------------------------------

constructor TDropMetaTarget.Create( AOwner: TComponent );
begin
  inherited Create( AOwner );
  FMetafile := TMetafile.Create;
  FFormatTymed := '::';
end;
// -----------------------------------------------------------------------------

destructor TDropMetaTarget.Destroy;
begin
  FMetafile.Free;
  inherited Destroy;
end;
// ----------------------------------------------------------------------------- 

function TDropMetaTarget.HasValidFormats: Boolean;
begin
  Result := (DataObject.QueryGetData(MetaFormatEtc) = S_OK) or
            (DataObject.QueryGetData(EnhFormatEtc) = S_OK) or
            (DataObject.QueryGetData(MetaFormatEtc2) = S_OK) or
            (DataObject.QueryGetData(EnhFormatEtc2) = S_OK);
end;
// ----------------------------------------------------------------------------- 

procedure TDropMetaTarget.ClearData;
begin
  FMetafile.Handle := 0;
end;
// ----------------------------------------------------------------------------- 

function TDropMetaTarget.DoGetData: Boolean;
var
  PMfp: PMetafilePict;
  Medium: TStgMedium;
begin
  Result := False;
  FFormatTymed := '::';
  //--------------------------------------------------------------------------
  if (DataObject.GetData(EnhFormatEtc, Medium) = S_OK) then
  begin
    FFormatTymed := 'CF_ENHMETAFILE::TYMED_ENHMF';
    try
      Result := (Medium.hEnhMetaFile <> 0);
      if Result = True then
        FMetaFile.Handle := CopyEnhMetafile(Medium.hEnhMetaFile, Nil);
    finally
      ReleaseStgMedium(Medium);
    end;
  end
  //--------------------------------------------------------------------------
  else if (DataObject.GetData(MetaFormatEtc, Medium) = S_OK) then
  begin
    FFormatTymed := 'CF_METAFILEPICT::TYMED_MFPICT';
    try
      PMfp := GlobalLock(Medium.hMetafilePict);
      try
        Result := (PMfp <> nil);
        if Result = True then
          FMetaFile.Handle := WMF2EMF(PMfp^);
      finally
        GlobalUnlock(Medium.hMetafilePict);
      end;
    finally
      ReleaseStgMedium(Medium);
    end;
  end
  //--------------------------------------------------------------------------
  else if (DataObject.GetData(MetaFormatEtc2, Medium) = S_OK) then
  begin
    FFormatTymed := 'CF_METAFILEPICT::TYMED_HGLOBAL';
    try
      PMfp := GlobalLock(Medium.hGlobal);
      try
        Result := (PMfp <> nil);
        if Result = True then
          FMetaFile.Handle := WMF2EMF(PMfp^);
      finally
        GlobalUnlock(Medium.hGlobal);
      end;
    finally
      ReleaseStgMedium(Medium);
    end;
  end
  //--------------------------------------------------------------------------
  else if (DataObject.GetData(EnhFormatEtc2, Medium) = S_OK) then
  begin
    FFormatTymed := 'CF_ENHMETAFILE::TYMED_HGLOBAL';
    try
      Result := (Medium.hGlobal <> 0);
      if Result = True then
        FMetaFile.Handle := CopyEnhMetafile(Medium.hGlobal, Nil);
    finally
      ReleaseStgMedium(Medium);
    end;
  end;
  //--------------------------------------------------------------------------
end;
// ---------------------------------------------------------------------------

end.

