unit DropPictureSource;

// -----------------------------------------------------------------------------
// Project:         Drag and Drop Component Suite
// Component Names: TDropPictureSource
// Module:          DropPictureSource
// Description:     Implements Dragging & Dropping of Graphics
//                  FROM your application to another.
// Version:         3.7
// Date:            22-JUL-1999
// Target:          Win32, Delphi 3 - Delphi 5, C++ Builder 3, C++ Builder 4
// Authors:         Angus Johnson,   ajohnson@rpi.net.au
//                  Anders Melander, anders@melander.dk
//                                   http://www.melander.dk
// Copyright        (C) 1997-99 Angus Johnson & Anders Melander
// Copyright        (C) 2011 seizo,  seizo@mars.dti.ne.jp
// -----------------------------------------------------------------------------

interface

uses
  Windows, Classes, SysUtils, Graphics, ClipBrd, ActiveX, DropSource;

type
  TDropPictureSource = class(TDropSource)
  private
    FPicture: TPicture;
    FBitmap: TBitmap;
    FMetafile: TMetafile;
    FFormatTymed: String;
    procedure SetPicture(const Value: TPicture);
  protected
    function DoGetData(const FormatEtcIn: TFormatEtc; out Medium: TStgMedium):HResult; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function CutOrCopyToClipboard: Boolean; override;
    property FormatTymed: String read FFormatTymed;
  published
    property Picture: TPicture read FPicture write SetPicture;
  end;

procedure Register;
function GetHGlobalDIBFromBitmap(Bitmap: TBitmap): HGlobal;

implementation

procedure Register;
begin
  RegisterComponents('DragDrop', [TDropPictureSource]);
end;

// -----------------------------------------------------------------------------
//     Miscellaneous Picture Function
// -----------------------------------------------------------------------------

function GetHGlobalDIBFromBitmap(Bitmap: TBitmap): HGlobal;
var
  Stream: TMemoryStream;
  DIB: Pointer;
  DIBSize: Integer;
begin
  Stream := TMemoryStream.Create;
  try
    // Write bitmap to a stream and extract the DIB data from it.
    Bitmap.SaveToStream(Stream);

    // Calculate size of DIB block.
    DIBSize := Stream.Size - SizeOf(TBitmapFileHeader);

    // Allocate memory for DIB data.
    Result := GlobalAlloc(GMEM_MOVEABLE or GMEM_ZEROINIT, DIBSize);
    if (Result = 0) then
      exit;

    DIB := GlobalLock(Result);
    if DIB = Nil then
    begin
      GlobalFree(Result);
      Result := 0;
    end
    else
    begin
      // Skip BMP file header.
      Stream.Seek(SizeOf(TBitmapFileHeader), soFromBeginning);
      // Transfer data from stream to global memory.
      if (Stream.Read(DIB^, DIBSize) <> DIBSize) then
      begin
        GlobalUnlock(Result);
        GlobalFree(Result);
        Result := 0;
      end
      else
        GlobalUnlock(Result);
    end;
  finally
    Stream.Free;
  end;
end;
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
//			TDropPictureSource
// -----------------------------------------------------------------------------

constructor TDropPictureSource.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPicture := TPicture.Create;
  FBitmap := TBitmap.Create;
  FMetafile := TMetafile.Create;
  DragTypes := [dtCopy]; // Default to Copy
  FFormatTymed := '::';

{ Add FORMATETC structure for Bitmap }
  AddFormatEtc(CF_DIB, Nil, DVASPECT_CONTENT, -1, TYMED_HGLOBAL);
  AddFormatEtc(CF_BITMAP, Nil, DVASPECT_CONTENT, -1, TYMED_GDI);
  AddFormatEtc(CF_PALETTE, Nil, DVASPECT_CONTENT, -1, TYMED_GDI);

{ Add FORMATETC structure for Metafile }
  AddFormatEtc(CF_METAFILEPICT, Nil, DVASPECT_CONTENT, -1, TYMED_MFPICT);
  AddFormatEtc(CF_ENHMETAFILE, Nil, DVASPECT_CONTENT, -1, TYMED_ENHMF);
  AddFormatEtc(CF_METAFILEPICT, Nil, DVASPECT_CONTENT, -1, TYMED_HGLOBAL);
  AddFormatEtc(CF_ENHMETAFILE, Nil, DVASPECT_CONTENT, -1, TYMED_HGLOBAL);
end;
// ----------------------------------------------------------------------------- 

destructor TDropPictureSource.Destroy;
begin
  FMetafile.Free;
  FBitmap.Free;
  FPicture.Free;
  inherited Destroy;
end;
// ----------------------------------------------------------------------------- 

procedure TDropPictureSource.SetPicture(const Value: TPicture);
begin
  FPicture.Graphic := Nil;
  FBitmap.Handle := 0;
  FMetafile.Handle := 0;
  FPicture.Assign(Value);
  if Value.Graphic is TBitmap then
    FBitmap.Assign(Value.Graphic)
  else if Value.Graphic is TMetafile then
    FMetafile.Assign(Value.Graphic)
  else
  begin
    FBitmap.Width := Value.Graphic.Width;
    FBitmap.Height := Value.Graphic.Height;
    FBitmap.Canvas.Draw(0, 0, Value.Graphic);
  end;
end;
// ----------------------------------------------------------------------------- 

function TDropPictureSource.CutOrCopyToClipboard: Boolean;
var
  Format: Word;
  Data: THandle;
  Palette: HPALETTE;
begin
  Result := False;
  if FPicture.Graphic.Empty then
    Exit;
  FPicture.SaveToClipboardFormat(Format, Data, Palette);
  try
    if Data = 0 then Exit;
    Clipboard.SetAsHandle(Format, Data);
    Result := True;
  except
    raise Exception.Create('Unable to copy Graphic to Clipboard.');
  end;
end;
// ----------------------------------------------------------------------------- 

function TDropPictureSource.DoGetData(const FormatEtcIn: TFormatEtc; out Medium: TStgMedium):HResult;
var
  fmt: WORD;
  pal: HPALETTE;
begin

  Medium.tymed := 0;
  Medium.UnkForRelease := Nil;
  Medium.hGlobal := 0;
  FFormatTymed := '::';
  //--------------------------------------------------------------------------

  if not FBitmap.Empty and
     (FormatEtcIn.cfFormat = CF_BITMAP) and
     (FormatEtcIn.dwAspect = DVASPECT_CONTENT) and
     (FormatEtcIn.tymed and TYMED_GDI <> 0) then
  begin
    FFormatTymed := 'CF_BITMAP::TYMED_GDI';
    try
      //This next line just gets a copy of the bitmap handle...
      FBitmap.SaveToClipboardFormat(fmt, THandle(Medium.hBitmap), pal);
      if pal <> 0 then
        DeleteObject(pal);
      if Medium.hBitmap <> 0 then
      begin
        Medium.tymed := TYMED_GDI;
        Result := S_OK;
      end
      else
        Result := E_OUTOFMEMORY;
    except
      Result := E_OUTOFMEMORY;
    end;
  end
  //-----------------------------------------------------------------------
  else if not FBitmap.Empty and
     (FormatEtcIn.cfFormat = CF_DIB) and
     (FormatEtcIn.dwAspect = DVASPECT_CONTENT) and
     (FormatEtcIn.tymed and TYMED_HGLOBAL <> 0) then
  begin
    FFormatTymed := 'CF_DIB::TYMED_HGLOBAL';
    try
      Medium.hGlobal := GetHGlobalDIBFromBitmap(FBitmap);
      if Medium.hGlobal <> 0 then
      begin
        Medium.tymed := TYMED_HGLOBAL;
        Result := S_OK
      end
      else
        Result := E_OUTOFMEMORY;
    except
      Result := E_OUTOFMEMORY;
    end;
  end
  //-----------------------------------------------------------------------
  else if not FBitmap.Empty and
     (FormatEtcIn.cfFormat = CF_PALETTE) and
     (FormatEtcIn.dwAspect = DVASPECT_CONTENT) and
     (FormatEtcIn.tymed and TYMED_GDI <> 0) then
  begin
    FFormatTymed := 'CF_PALETTE::TYMED_GDI';
    try
      Medium.hBitmap := CopyPalette(FBitmap.Palette);
      if Medium.hBitmap <> 0 then
      begin
        Medium.tymed := TYMED_GDI;
        Result := S_OK;
      end
      else
        Result := E_OUTOFMEMORY;
    except
      Result := E_OUTOFMEMORY;
    end;
  end
  //-----------------------------------------------------------------------
  else if not FMetafile.Empty and
     (FormatEtcIn.cfFormat = CF_ENHMETAFILE) and
     (FormatEtcIn.dwAspect = DVASPECT_CONTENT) and
     (FormatEtcIn.tymed and TYMED_ENHMF <> 0) then
  begin
    FFormatTymed := 'CF_ENHMETAFILE::TYMED_ENHMF';
    try
      //This next line just gets a copy of the metafile handle...
      FMetafile.SaveToClipboardFormat(fmt, THandle(Medium.hEnhMetaFile), pal);
      if pal <> 0 then
        DeleteObject(pal);
      if Medium.hEnhMetaFile <> 0 then
      begin
        Medium.tymed := TYMED_ENHMF;
        Result := S_OK;
      end
      else
        Result := E_OUTOFMEMORY;
    except
      Result := E_OUTOFMEMORY;
    end;
  end
  //-----------------------------------------------------------------------
  else if not FMetafile.Empty and
     (FormatEtcIn.cfFormat = CF_METAFILEPICT) and
     (FormatEtcIn.dwAspect = DVASPECT_CONTENT) and
     (FormatEtcIn.tymed and TYMED_MFPICT <> 0) then
  begin
    FFormatTymed := 'CF_METAFILEPICT::TYMED_MFPICT';
    try
      //This next line just gets a copy of the metafile handle...
      FMetafile.SaveToClipboardFormat(fmt, THandle(Medium.hMetafilePict), pal);
      if pal <> 0 then
        DeleteObject(pal);
      if Medium.hMetafilePict <> 0 then
      begin
        Medium.tymed := TYMED_MFPICT;
        Result := S_OK;
      end
      else
        Result := E_OUTOFMEMORY;
    except
      Result := E_OUTOFMEMORY;
    end;
  end
  //-----------------------------------------------------------------------
  else if not FMetafile.Empty and
     (FormatEtcIn.cfFormat = CF_METAFILEPICT) and
     (FormatEtcIn.dwAspect = DVASPECT_CONTENT) and
     (FormatEtcIn.tymed and TYMED_HGLOBAL <> 0) then
  begin
    FFormatTymed := 'CF_METAFILEPICT::TYMED_HGLOBAL';
    try
      //This next line just gets a copy of the metafile handle...
      FMetafile.SaveToClipboardFormat(fmt, THandle(Medium.hGlobal), pal);
      if pal <> 0 then
        DeleteObject(pal);
      if Medium.hGlobal <> 0 then
      begin
        Medium.tymed := TYMED_HGLOBAL;
        Result := S_OK;
      end
      else
        Result := E_OUTOFMEMORY;
    except
      Result := E_OUTOFMEMORY;
    end;
  end
  //-----------------------------------------------------------------------
  else if not FMetafile.Empty and
     (FormatEtcIn.cfFormat = CF_ENHMETAFILE) and
     (FormatEtcIn.dwAspect = DVASPECT_CONTENT) and
     (FormatEtcIn.tymed and TYMED_HGLOBAL <> 0) then
  begin
    FFormatTymed := 'CF_ENHMETAFILE::TYMED_HGLOBAL';
    try
      //This next line just gets a copy of the metafile handle...
      FMetafile.SaveToClipboardFormat(fmt, THandle(Medium.hGlobal), pal);
      if pal <> 0 then
        DeleteObject(pal);
      if Medium.hGlobal <> 0 then
      begin
        Medium.tymed := TYMED_HGLOBAL;
        Result := S_OK;
      end
      else
        Result := E_OUTOFMEMORY;
    except
      Result := E_OUTOFMEMORY;
    end;
  end
  //-----------------------------------------------------------------------
  else
    Result := DV_E_FORMATETC;
//-------------------------------------------------------------------------
end;
// ----------------------------------------------------------------------------

end.

