unit DropPictureTarget;

// -----------------------------------------------------------------------------
// Project:         Drag and Drop Component Suite
// Component Names: TDropPictureTarget
// Module:          DropPictureTarget
// Description:     Implements Dragging & Dropping of Graphics
//                  FROM your application to another.
// Version:         3.7
// Date:            22-JUL-1999
// Target:          Win32, Delphi 3 - Delphi 5, C++ Builder 3, C++ Builder 4
// Authors:         Angus Johnson,   ajohnson@rpi.net.au
//                  Anders Melander, anders@melander.dk
//                                   http://www.melander.dk
// Copyright        (C) 1997-99 Angus Johnson & Anders Melander
// Copyright        (C) 2011 seizo,  seizo@mars.dti.ne.jp
// -----------------------------------------------------------------------------

interface

uses
  Windows, Classes, SysUtils, Graphics, ActiveX, DropSource, DropTarget;

type
  TDropPictureTarget = class(TDropTarget)
  private
    FPicture: TPicture;
    FBitmap: TBitmap;
    FMetafile: TMetafile;
    FFormatTymed: String;
  protected
    procedure ClearData; override;
    function DoGetData: Boolean; override;
    function HasValidFormats: Boolean; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Picture: TPicture read FPicture;
    property FormatTymed: String read FFormatTymed;
  end;

procedure Register;
procedure CopyDIBToBitmap(Bitmap: TBitmap; BitmapInfo: PBitmapInfo; DIBSize: Integer);
function WMF2EMF(const Mfp: TMetaFilePict): hEnhMetaFile;

implementation

const
{ FORMATETC structure for Bitmap }
  DIBFormatEtc: TFormatEtc = (cfFormat: CF_DIB;
    ptd: Nil; dwAspect: DVASPECT_CONTENT; lindex: -1; tymed: TYMED_HGLOBAL);
  BMPFormatEtc: TFormatEtc = (cfFormat: CF_BITMAP;
    ptd: Nil; dwAspect: DVASPECT_CONTENT; lindex: -1; tymed: TYMED_GDI);
  PalFormatEtc: TFormatEtc = (cfFormat: CF_PALETTE;
    ptd: Nil; dwAspect: DVASPECT_CONTENT; lindex: -1; tymed: TYMED_GDI);

{ FORMATETC structure for Metafile }
  MetaFormatEtc: TFormatEtc = (cfFormat: CF_METAFILEPICT;
    ptd: Nil; dwAspect: DVASPECT_CONTENT; lindex: -1; tymed: TYMED_MFPICT);
  EnhFormatEtc: TFormatEtc = (cfFormat: CF_ENHMETAFILE;
    ptd: Nil; dwAspect: DVASPECT_CONTENT; lindex: -1; tymed: TYMED_ENHMF);
  MetaFormatEtc2: TFormatEtc = (cfFormat: CF_METAFILEPICT;
    ptd: Nil; dwAspect: DVASPECT_CONTENT; lindex: -1; tymed: TYMED_HGLOBAL);
  EnhFormatEtc2: TFormatEtc = (cfFormat: CF_ENHMETAFILE;
    ptd: Nil; dwAspect: DVASPECT_CONTENT; lindex: -1; tymed: TYMED_HGLOBAL);

procedure Register;
begin
  RegisterComponents('DragDrop', [TDropPictureTarget]);
end;

// -----------------------------------------------------------------------------
//	Miscellaneous Picture Functions
// -----------------------------------------------------------------------------

procedure CopyDIBToBitmap(Bitmap: TBitmap; BitmapInfo: PBitmapInfo; DIBSize: Integer);
var
  BitmapFileHeader: TBitmapFileHeader;
  FileSize: Integer;
  InfoSize: Integer;
  Stream: TMemoryStream;
begin
  // Write DIB to a stream in the BMP file format
  Stream := TMemoryStream.Create;
  try
    FileSize := SizeOf(TBitmapFileHeader) + DIBSize;
    InfoSize := SizeOf(TBitmapInfoHeader);
    if (BitmapInfo^.bmiHeader.biBitCount > 8) then
    begin
      if ((BitmapInfo^.bmiHeader.biCompression and BI_BITFIELDS) <> 0) then
        Inc(InfoSize, 12);
    end else
      Inc(InfoSize, SizeOf(TRGBQuad) * (1 shl BitmapInfo^.bmiHeader.biBitCount));
    Stream.SetSize(FileSize);
    // Initialize file header
    FillChar(BitmapFileHeader, SizeOf(TBitmapFileHeader), 0);
    with BitmapFileHeader do
    begin
      bfType := $4D42; // 'BM' = Windows BMP signature
      bfSize := FileSize; // File size (not needed)
      bfOffBits := SizeOf(TBitmapFileHeader) + InfoSize; // Offset of pixel data
    end;
    // Save file header
    Stream.Write(BitmapFileHeader, SizeOf(TBitmapFileHeader));
    // Save TBitmapInfo structure and pixel data
    Stream.Write(BitmapInfo^, DIBSize);

    // Rewind and load bitmap from stream
    Stream.Position := 0;
    Bitmap.LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

// -----------------------------------------------------------------------------

function WMF2EMF(const Mfp: TMetaFilePict): hEnhMetaFile;
var
  Bits: Pointer;
  Length: UINT;
  RefDC: HDC;
begin
  Length := GetMetaFileBitsEx(Mfp.hMF, 0, Nil);
  if Length = 0 then
    RaiseLastWin32Error;
  GetMem(Bits, Length);
  try
    if GetMetaFileBitsEx(Mfp.hMF, Length, Bits) < Length then
      RaiseLastWin32Error;
    RefDC := GetDC(0);
    try
        Result := SetWinMetaFileBits(Length, Bits, RefDC, Mfp);
    finally
      ReleaseDC(0, RefDC);
    end;
    if Result = 0 then
      RaiseLastWin32Error;

  finally
    FreeMem(Bits);
  end;
end;
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
//			TDropPictureTarget
// -----------------------------------------------------------------------------

constructor TDropPictureTarget.Create( AOwner: TComponent );
begin
  inherited Create( AOwner );
  FPicture := TPicture.Create;
  FBitmap := TBitmap.Create;
  FMetafile := TMetafile.Create;
  FFormatTymed := '::';
end;
// -----------------------------------------------------------------------------

destructor TDropPictureTarget.Destroy;
begin
  FMetafile.Free;
  FBitmap.Free;
  FPicture.Free;
  inherited Destroy;
end;
// ----------------------------------------------------------------------------- 

function TDropPictureTarget.HasValidFormats: Boolean;
begin
  Result := (DataObject.QueryGetData(DIBFormatEtc) = S_OK) or
            (DataObject.QueryGetData(BMPFormatEtc) = S_OK) or
            (DataObject.QueryGetData(MetaFormatEtc) = S_OK) or
            (DataObject.QueryGetData(EnhFormatEtc) = S_OK) or
            (DataObject.QueryGetData(MetaFormatEtc2) = S_OK) or
            (DataObject.QueryGetData(EnhFormatEtc2) = S_OK) or
            (DataObject.QueryGetData(PalFormatEtc) = S_OK);
end;
// ----------------------------------------------------------------------------- 

procedure TDropPictureTarget.ClearData;
begin
  FPicture.Graphic := Nil;
end;
// ----------------------------------------------------------------------------- 

function TDropPictureTarget.DoGetData: Boolean;
var
  Medium, Medium2: TStgMedium;
  DIBData: Pointer;
  Pal: HPALETTE;
  PMfp: PMetafilePict;
begin
  Result := False;
  FFormatTymed := '::';
  //--------------------------------------------------------------------------
  if (DataObject.GetData(BMPFormatEtc, Medium) = S_OK) then
  begin
    FFormatTymed := 'CF_BITMAP::TYMED_GDI';
    try
      Result := (Medium.hBitmap <> 0);
      if Result = True then
      begin
        if (DataObject.GetData(PalFormatEtc, Medium2) = S_OK) then
        begin
          FFormatTymed := FFormatTymed + '  CF_PALETTE::TYMED_GDI';
          try
            Pal := CopyPalette(Medium2.hBitmap);
            FBitmap.LoadFromClipboardFormat(CF_BITMAP, Medium.hBitmap, Pal);
            FPicture.Assign(FBitmap);
          finally
            ReleaseStgMedium(Medium2);
          end;
        end
        else
        begin
          FBitmap.LoadFromClipboardFormat(CF_BITMAP, Medium.hBitmap, 0);
          FPicture.Assign(FBitmap);
        end;
      end;
    finally
      FBitmap.Handle := 0;
      ReleaseStgMedium(Medium);
    end;
  end
  //--------------------------------------------------------------------------
  else if (DataObject.GetData(DIBFormatEtc, Medium) = S_OK) then
  begin
    FFormatTymed := 'CF_DIB::TYMED_HGLOBAL';
    try
      DIBData := GlobalLock(Medium.hGlobal);
      try
        Result := (DIBData <> Nil);
        if Result = True then
        begin
          CopyDIBToBitmap(FBitmap, DIBData, GlobalSize(Medium.hGlobal));
          FPicture.Assign(FBitmap);
        end;
      finally
        GlobalUnlock(Medium.hGlobal);
      end;
    finally
      FBitmap.Handle := 0;
      ReleaseStgMedium(Medium);
    end;
  end
  //--------------------------------------------------------------------------
  else if (DataObject.GetData(PalFormatEtc, Medium) = S_OK) then
  begin
    FFormatTymed := 'CF_PALETTE::TYMED_GDI';
    try
      Result := (Medium.hBitmap <> 0);
      if Result = True then
      begin
        FBitmap.Handle := CopyPalette(Medium.hBitmap);
        FPicture.Assign(FBitmap);
      end;
    finally
      FBitmap.Handle := 0;
      ReleaseStgMedium(Medium);
    end;
  end
  //--------------------------------------------------------------------------
  else if (DataObject.GetData(EnhFormatEtc, Medium) = S_OK) then
  begin
    FFormatTymed := 'CF_ENHMETAFILE::TYMED_ENHMF';
    try
      Result := (Medium.hEnhMetaFile <> 0);
      if Result = True then
      begin
        FMetafile.Handle := CopyEnhMetafile(Medium.hEnhMetaFile, Nil);
        FPicture.Assign(FMetafile);
      end;
    finally
      FMetafile.Handle := 0;
      ReleaseStgMedium(Medium);
    end;
  end
  //--------------------------------------------------------------------------
  else if (DataObject.GetData(MetaFormatEtc, Medium) = S_OK) then
  begin
    FFormatTymed := 'CF_METAFILEPICT::TYMED_MFPICT';
    try
      PMfp := GlobalLock(Medium.hMetafilePict);
      try
        Result := (PMfp <> nil);
        if Result = True then
        begin
          FMetafile.Handle := WMF2EMF(PMfp^);
          FPicture.Assign(FMetafile);
        end;
      finally
        GlobalUnlock(Medium.hMetafilePict);
      end;
    finally
      FMetafile.Handle := 0;
      ReleaseStgMedium(Medium);
    end;
  end
  //--------------------------------------------------------------------------
  else if (DataObject.GetData(MetaFormatEtc2, Medium) = S_OK) then
  begin
    FFormatTymed := 'CF_METAFILEPICT::TYMED_HGLOBAL';
    try
      PMfp := GlobalLock(Medium.hGlobal);
      try
        Result := (PMfp <> nil);
        if Result = True then
        begin
          FMetafile.Handle := WMF2EMF(PMfp^);
          FPicture.Assign(FMetafile);
        end;
      finally
        GlobalUnlock(Medium.hGlobal);
      end;
    finally
      FMetafile.Handle := 0;
      ReleaseStgMedium(Medium);
    end;
  end
  //--------------------------------------------------------------------------
  else if (DataObject.GetData(EnhFormatEtc2, Medium) = S_OK) then
  begin
    FFormatTymed := 'CF_ENHMETAFILE::TYMED_HGLOBAL';
    try
      Result := (Medium.hGlobal <> 0);
      if Result = True then
      begin
        FMetafile.Handle := CopyEnhMetafile(Medium.hGlobal, Nil);
        FPicture.Assign(FMetafile);
      end;
    finally
      FMetafile.Handle := 0;
      ReleaseStgMedium(Medium);
    end;
  end;
  //--------------------------------------------------------------------------
end;
// -----------------------------------------------------------------------------

end.

