unit DropBMPSource;

// -----------------------------------------------------------------------------
// Project:         Drag and Drop Component Suite
// Component Names: TDropBMPSource
// Module:          DropBMPSource
// Description:     Implements Dragging & Dropping of Bitmaps
//                  FROM your application to another.
// Version:         3.7
// Date:            22-JUL-1999
// Target:          Win32, Delphi 3 - Delphi 5, C++ Builder 3, C++ Builder 4
// Authors:         Angus Johnson,   ajohnson@rpi.net.au
//                  Anders Melander, anders@melander.dk
//                                   http://www.melander.dk
// Copyright        (C) 1997-99 Angus Johnson & Anders Melander
// Modified             2011 seizo,  seizo@mars.dti.ne.jp
// -----------------------------------------------------------------------------
// Acknowledgements:
// Thanks to Dieter Steinwedel for some help with DIBs.
// http://godard.oec.uni-osnabrueck.de/student_home/dsteinwe/delphi/DietersDelphiSite.htm
// -----------------------------------------------------------------------------

interface

uses
  Windows, Classes, SysUtils, Graphics, ClipBrd, ActiveX, DropSource;

type
  TDropBMPSource = class(TDropSource)
  private
    FBitmap: TBitmap;
    FFormatTymed: String;
    procedure SetBitmap(const Value: TBitmap);
  protected
    function DoGetData(const FormatEtcIn: TFormatEtc; out Medium: TStgMedium):HResult; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function CutOrCopyToClipboard: Boolean; override;
    property FormatTymed: String read FFormatTymed;
  published
    property Bitmap: TBitmap read FBitmap write SetBitmap;
  end;

procedure Register;
function GetHGlobalDIBFromBitmap(Bitmap: TBitmap): HGlobal;

implementation

procedure Register;
begin
  RegisterComponents('DragDrop', [TDropBMPSource]);
end;

// -----------------------------------------------------------------------------
//     Miscellaneous DIB Function
// -----------------------------------------------------------------------------

function GetHGlobalDIBFromBitmap(Bitmap: TBitmap): HGlobal;
var
  Stream: TMemoryStream;
  DIB: Pointer;
  DIBSize: Integer;
begin
  Stream := TMemoryStream.Create;
  try
    // Write bitmap to a stream and extract the DIB data from it.
    Bitmap.SaveToStream(Stream);

    // Calculate size of DIB block.
    DIBSize := Stream.Size - SizeOf(TBitmapFileHeader);

    // Allocate memory for DIB data.
    Result := GlobalAlloc(GMEM_MOVEABLE or GMEM_ZEROINIT, DIBSize);
    if (Result = 0) then
      exit;

    DIB := GlobalLock(Result);
    if DIB = Nil then
    begin
      GlobalFree(Result);
      Result := 0;
    end
    else
    begin
      // Skip BMP file header.
      Stream.Seek(SizeOf(TBitmapFileHeader), soFromBeginning);
      // Transfer data from stream to global memory.
      if (Stream.Read(DIB^, DIBSize) <> DIBSize) then
      begin
        GlobalUnlock(Result);
        GlobalFree(Result);
        Result := 0;
      end
      else
        GlobalUnlock(Result);
    end;
  finally
    Stream.Free;
  end;
end;

// -----------------------------------------------------------------------------
//			TDropBMPSource
// -----------------------------------------------------------------------------

constructor TDropBMPSource.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FBitmap := TBitmap.Create;
  DragTypes := [dtCopy]; // Default to Copy
  FFormatTymed := '::';

  AddFormatEtc(CF_DIB, Nil, DVASPECT_CONTENT, -1, TYMED_HGLOBAL);
  AddFormatEtc(CF_BITMAP, Nil, DVASPECT_CONTENT, -1, TYMED_GDI);
  AddFormatEtc(CF_PALETTE, Nil, DVASPECT_CONTENT, -1, TYMED_GDI);
end;
// ----------------------------------------------------------------------------- 

destructor TDropBMPSource.destroy;
begin
  FBitmap.Free;
  inherited Destroy;
end;
// ----------------------------------------------------------------------------- 

procedure TDropBMPSource.SetBitmap(const Value: TBitmap);
begin
  FBitmap.Assign(Value);
end;
// ----------------------------------------------------------------------------- 

function TDropBMPSource.CutOrCopyToClipboard: Boolean;
var
  BFormat: Word;
  AData: THandle;
  APalette: HPALETTE;
begin
  Result := False;
  if FBitmap.Empty then Exit;
  FBitmap.SaveToClipboardFormat(BFormat, AData, APalette);
  try
    if AData = 0 then Exit;
    Clipboard.SetAsHandle(BFormat, AData);
    Result := True;
  except
    raise Exception.Create('Unable to copy Bitmap to Clipboard.');
  end;
end;
// ----------------------------------------------------------------------------- 

function TDropBMPSource.DoGetData(const FormatEtcIn: TFormatEtc; out Medium: TStgMedium):HResult;
var
  fmt: WORD;
  pal: HPALETTE;
begin

  Medium.tymed := 0;
  Medium.UnkForRelease := Nil;
  Medium.hGlobal := 0;
  FFormatTymed := '::';
  //--------------------------------------------------------------------------
  if not FBitmap.Empty and
     (FormatEtcIn.cfFormat = CF_BITMAP) and
     (FormatEtcIn.dwAspect = DVASPECT_CONTENT) and
     (FormatEtcIn.tymed and TYMED_GDI <> 0) then
  begin
    FFormatTymed := 'CF_BITMAP::TYMED_GDI';
    try
      //This next line just gets a copy of the bitmap handle...
      FBitmap.SaveToClipboardFormat(fmt, THandle(Medium.hBitmap), pal);
      if pal <> 0 then
        DeleteObject(pal);
      if Medium.hBitmap <> 0 then
      begin
        Medium.tymed := TYMED_GDI;
        Result := S_OK;
      end
      else
        Result := E_OUTOFMEMORY;
    except
      Result := E_OUTOFMEMORY;
    end;
  end
  //--------------------------------------------------------------------------
  else if not FBitmap.Empty and
     (FormatEtcIn.cfFormat = CF_DIB) and
     (FormatEtcIn.dwAspect = DVASPECT_CONTENT) and
     (FormatEtcIn.tymed and TYMED_HGLOBAL <> 0) then
  begin
    FFormatTymed := 'CF_DIB::TYMED_HGLOBAL';
    try
      Medium.hGlobal := GetHGlobalDIBFromBitmap(FBitmap);
      if Medium.hGlobal <> 0 then
      begin
        Medium.tymed := TYMED_HGLOBAL;
        Result := S_OK
      end else
        Result := E_OUTOFMEMORY;
    except
      Result := E_OUTOFMEMORY;
    end;
  end
  //--------------------------------------------------------------------------
  else if not FBitmap.Empty and
     (FormatEtcIn.cfFormat = CF_PALETTE) and
     (FormatEtcIn.dwAspect = DVASPECT_CONTENT) and
     (FormatEtcIn.tymed and TYMED_GDI <> 0) then
  begin
    FFormatTymed := 'CF_PALETTE::TYMED_GDI';
    try
      Medium.hBitmap := CopyPalette(FBitmap.Palette);
      if Medium.hBitmap <> 0 then
      begin
        Medium.tymed := TYMED_GDI;
        Result := S_OK;
      end
      else
        Result := E_OUTOFMEMORY;
    except
      Result := E_OUTOFMEMORY;
    end;
  end
  //--------------------------------------------------------------------------
  else
    Result := DV_E_FORMATETC;
end;
// ----------------------------------------------------------------------------- 

end.

