unit DropBMPTarget;

// -----------------------------------------------------------------------------
// Project:         Drag and Drop Component Suite
// Component Names: TDropBMPTarget
// Module:          DropBMPTarget
// Description:     Implements Dragging & Dropping of Bitmaps
//                  FROM your application to another.
// Version:         3.7
// Date:            22-JUL-1999
// Target:          Win32, Delphi 3 - Delphi 5, C++ Builder 3, C++ Builder 4
// Authors:         Angus Johnson,   ajohnson@rpi.net.au
//                  Anders Melander, anders@melander.dk
//                                   http://www.melander.dk
// Copyright        (C) 1997-99 Angus Johnson & Anders Melander
// Modified             2011 seizo,  seizo@mars.dti.ne.jp
// -----------------------------------------------------------------------------

interface

uses
  Windows, Classes, Graphics, ActiveX, DropSource, DropTarget;

type
  TDropBMPTarget = class(TDropTarget)
  private
    FBitmap: TBitmap;
    FFormatTymed: String;
  protected
    procedure ClearData; override;
    function DoGetData: Boolean; override;
    function HasValidFormats: Boolean; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Bitmap: TBitmap read FBitmap;
    property FormatTymed: String read FFormatTymed;
  end;

procedure Register;
procedure CopyDIBToBitmap(Bitmap: TBitmap; BitmapInfo: PBitmapInfo; DIBSize: Integer);

implementation

const
  DIBFormatEtc: TFormatEtc = (cfFormat: CF_DIB;
    ptd: Nil; dwAspect: DVASPECT_CONTENT; lindex: -1; tymed: TYMED_HGLOBAL);
  BMPFormatEtc: TFormatEtc = (cfFormat: CF_BITMAP;
    ptd: Nil; dwAspect: DVASPECT_CONTENT; lindex: -1; tymed: TYMED_GDI);
  PalFormatEtc: TFormatEtc = (cfFormat: CF_PALETTE;
    ptd: Nil; dwAspect: DVASPECT_CONTENT; lindex: -1; tymed: TYMED_GDI);

procedure Register;
begin
  RegisterComponents('DragDrop', [TDropBMPTarget]);
end;

// -----------------------------------------------------------------------------
//	Miscellaneous DIB Function
// -----------------------------------------------------------------------------

procedure CopyDIBToBitmap(Bitmap: TBitmap; BitmapInfo: PBitmapInfo; DIBSize: Integer);
var
  BitmapFileHeader: TBitmapFileHeader;
  FileSize: Integer;
  InfoSize: Integer;
  Stream: TMemoryStream;
begin
  // Write DIB to a stream in the BMP file format
  Stream := TMemoryStream.Create;
  try
    FileSize := SizeOf(TBitmapFileHeader) + DIBSize;
    InfoSize := SizeOf(TBitmapInfoHeader);
    if (BitmapInfo^.bmiHeader.biBitCount > 8) then
    begin
      if ((BitmapInfo^.bmiHeader.biCompression and BI_BITFIELDS) <> 0) then
        Inc(InfoSize, 12);
    end else
      Inc(InfoSize, SizeOf(TRGBQuad) * (1 shl BitmapInfo^.bmiHeader.biBitCount));
    Stream.SetSize(FileSize);
    // Initialize file header
    FillChar(BitmapFileHeader, SizeOf(TBitmapFileHeader), 0);
    with BitmapFileHeader do
    begin
      bfType := $4D42; // 'BM' = Windows BMP signature
      bfSize := FileSize; // File size (not needed)
      bfOffBits := SizeOf(TBitmapFileHeader) + InfoSize; // Offset of pixel data
    end;
    // Save file header
    Stream.Write(BitmapFileHeader, SizeOf(TBitmapFileHeader));
    // Save TBitmapInfo structure and pixel data
    Stream.Write(BitmapInfo^, DIBSize);

    // Rewind and load bitmap from stream
    Stream.Position := 0;
    Bitmap.LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

// -----------------------------------------------------------------------------
//			TDropBMPTarget
// -----------------------------------------------------------------------------

constructor TDropBMPTarget.Create( AOwner: TComponent );
begin
  inherited Create( AOwner );
  FBitmap := TBitmap.Create;
  FFormatTymed := '::';
end;
// -----------------------------------------------------------------------------

destructor TDropBMPTarget.Destroy;
begin
  FBitmap.Free;
  inherited Destroy;
end;
// ----------------------------------------------------------------------------- 

function TDropBMPTarget.HasValidFormats: Boolean;
begin
  Result := (DataObject.QueryGetData(DIBFormatEtc) = S_OK) or
                (DataObject.QueryGetData(BMPFormatEtc) = S_OK);
end;
// ----------------------------------------------------------------------------- 

procedure TDropBMPTarget.ClearData;
begin
  FBitmap.Handle := 0;
end;
// ----------------------------------------------------------------------------- 

function TDropBMPTarget.DoGetData: Boolean;
var
  Medium, Medium2: TStgMedium;
  DIBData: Pointer;
  Pal: HPALETTE;
begin
  Result := False;
  FFormatTymed := '::';
  //--------------------------------------------------------------------------
  if (DataObject.GetData(BMPFormatEtc, Medium) = S_OK) then
  begin
    FFormatTymed := 'CF_BITMAP::TYMED_GDI';
    try
      Result := (Medium.hBitmap <> 0);
      if Result = True then
      begin
        if (DataObject.GetData(PalFormatEtc, Medium2) = S_OK) then
        begin
          FFormatTymed := FFormatTymed + '  CF_PALETTE::TYMED_GDI';
          try
            Pal := CopyPalette(Medium2.hBitmap);
            FBitmap.LoadFromClipboardFormat(CF_BITMAP, Medium.hBitmap, Pal);
          finally
            ReleaseStgMedium(Medium2);
          end;
        end
        else
          FBitmap.LoadFromClipboardFormat(CF_BITMAP, Medium.hBitmap, 0);
      end;
    finally
      ReleaseStgMedium(Medium);
    end;
  end
  //--------------------------------------------------------------------------
  else if (DataObject.GetData(DIBFormatEtc, Medium) = S_OK) then
  begin
    FFormatTymed := 'CF_DIB::TYMED_HGLOBAL';
    try
      DIBData := GlobalLock(Medium.hGlobal);
      try
        Result := (DIBData <> Nil);
        if Result = True then
          CopyDIBToBitmap(FBitmap, DIBData, GlobalSize(Medium.hGlobal));
      finally
        GlobalUnlock(Medium.hGlobal);
      end;
    finally
      ReleaseStgMedium(Medium);
    end;
  end
  //--------------------------------------------------------------------------
  else if (DataObject.GetData(PalFormatEtc, Medium) = S_OK) then
  begin
    FFormatTymed := 'CF_PALETTE::TYMED_GDI';
    try
      Result := (Medium.hBitmap <> 0);
      if Result = True then
        FBitmap.Handle := CopyPalette(Medium.hBitmap);
    finally
      ReleaseStgMedium(Medium);
    end;
  end;
  //--------------------------------------------------------------------------
end;
// -----------------------------------------------------------------------------

end.

